import sys
from collections import defaultdict
from decimal import Decimal
from decimal import InvalidOperation as DecimalInvalidOperation
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Sequence,
    Tuple,
    TypeVar,
    Union,
    overload,
)

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from pnq._itertools.op import MAP_ASSIGN_OP, TH_ASSIGN_OP, TH_ROUND
from pnq.exceptions import (
    DuplicateElementError,
    MustError,
    MustTypeError,
    NoElementError,
    NotFoundError,
    NotOneElementError,
)

if TYPE_CHECKING:
    from _typeshed import SupportsKeysAndGetItem


T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


class EmptyLambda:
    def __call__(self, x):
        return x

    def __str__(self):
        return "lambda x: x"


lambda_empty = EmptyLambda()
# lambda_empty = lambda x: x
# lambda_empty.__str__ = lambda: "lambda x: x"


def name_as(name):
    def wrapper(func):
        func.__name__ = name
        return func

    return wrapper


marked = []


def mark(func):
    marked.append(func)
    return func


def name_as(name):
    def wrapper(func):
        func.__name__ = name
        return func

    return wrapper


@name_as("list")
def __list(iterable: Iterable[T]) -> List[T]:
    return list(iterable)


@overload
def __dict(mapping: "SupportsKeysAndGetItem[K, V]") -> Dict[K, V]:
    return Dict[K, V](mapping)


@overload
def __dict(iterable: Iterable[Tuple[K, V]]) -> Dict[K, V]:
    return Dict[K, V](iterable)


@name_as("dict")
def __dict(iterable):
    return Dict[K, V](iterable)


###########################################
# generator
###########################################
@mark
@name_as("iter")
def __iter(self, selector=None):
    """イテラブルまたはマッピングからイテレータを取得します。
    マッピングの場合は、キーバリューのタプルを返すイテレータを取得します。

    Args:

    * self: イテレータを取得するイテラブルまたはマッピング
    * selector: 取得したイテレータから任意の要素を抽出する関数

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.iter([1, 2]).to(list)
    [1, 2]
    >>> pnq.iter([{"id": 1, "name": "bob"}]).to(list)
    [("id", 1), ("name", "bob")]
    >>> pnq.iter([{"id": 1, "name": "bob"}], lambda x: x[1]).to(list)
    [1, "bob"]
    ```
    """


def __map_nullable(self, selector):
    if selector is None:
        return self
    else:
        return map(selector, self)


@mark
def __next(self):
    raise NotImplementedError()


@mark
def value(*args, **kwargs):
    """１つの要素を返すイテレータを生成します。
    位置引数を与えた場合は与えた値かタプルになり、キーワード引数を与えた場合は辞書になります。

    Usage:
    ```
    >>> pnq.value(1).to(list)
    [1]
    >>> pnq.value("a", "b").to(list)
    [("a", "b")]
    >>> pnq.value(name="test").to(list)
    [{"naem": "test"}]
    ```
    """
    if args and kwargs:
        raise ValueError("value() can't accept both positional and keyword arguments")

    if kwargs:
        val = kwargs

    elif len(args) > 1:
        val = args

    elif len(args) == 1:
        val = args[0]

    else:
        raise NotImplementedError()

    yield val


@mark
def infinite(func, *args, **kwargs):
    """渡した関数を無限に実行するイテレータを生成します。
    無限に繰り返されるため、`take`等で終了条件を設定するように注意してください。

    Args:

    * func(args, kwargs): 無限に実行する関数
    * args: 関数に渡す位置引数
    * kwargs: 関数に渡すキーワード引数

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.infinite(datetime.now).take(1).to(list)
    [datetime.datetime(2021, 9, 10, 3, 57, 54, 402467)]
    >>> pnq.infinite(datetime, 2020, 1, day=2).take(1).to(list)
    [datetime.datetime(2010, 1, 2, 0, 0)]
    ```
    """
    while True:
        yield func(*args, **kwargs)


@mark
def repeat(value):
    """同じ値を繰り返す無限イテレータを生成します。

    Args:

    * value: 繰り返す値

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.repeat(5).take(3).to(list)
    [5, 5, 5]
    ```
    """
    from itertools import repeat

    yield from repeat(value)


@mark
def count(start=0, step=1):
    """連続した値を無限に返すイテレータを生成します。
    無限に繰り返されるため、`take`等で終了条件を設定するように注意してください。

    Args:

    * start: 開始値
    * step: 増分

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.count().take(3).to(list)
    [0, 1, 2]
    >>> pnq.count(1, 2).take(3).to(list)
    [1, 3, 5]
    ```
    """
    from itertools import count

    yield from count(start, step)


@mark
def cycle(iterable, repeat=None):
    """イテラブルが返す値を無限に繰り返すイテレータを生成します。

    Args:

    * iterable: 繰り返すイテラブル
    * repeat: 繰り返す回数。Noneの場合は無限に繰り返します。

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.cycle([1,2,3]).take(4).to(list)
    [1, 2, 3, 1]
    >>> pnq.cycle([1,2,3], repeat=2).to(list)
    [1, 2, 3, 1, 2, 3]
    ```
    """
    from itertools import cycle

    yield from cycle(iterable, repeat)


@mark
@name_as("range")
def __range(*args, **kwargs):
    """指定した開始数と終了数までの連続した値を返すイテレータを生成します。

    Args:

    * stop: 終了数

    Args:

    * start: 開始数
    * stop: 終了数
    * step: 増分

    Returns: 取得したイテレータを内包するクエリ

    Usage:
    ```
    >>> pnq.range(5).to(list)
    [0, 1, 2, 3, 4]
    >>> pnq.range(0, 3, 2).to(list)
    [0, 2, 4]
    ```
    """
    yield from range(*args, **kwargs)


###########################################
# mapping
###########################################


@mark
def map(self, selector, unpack=""):
    """シーケンスの各要素を新しいフォームに射影します。
    str関数を渡した場合、`None`は`""`を返します（Pythonの標準動作は`"None"`を返します）。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素に対する変換関数
    * unpack: 引数をどのように展開するか指定する
        - `""`: 展開せずにそのまま値を渡す（デフォルト）
        - `"\*"`: 位置引数として展開する
        - `"\*\*"`: キーワード引数として展開する
        - `"\*\*\*"`: 位置引数とキーワード引数を展開する（pnq.Argumentsインスタンスに対して使用できます）

    Usage:
    ```
    >>> pnq.query([1]).map(lambda x: x * 2).to(list)
    [2]
    >>> pnq.query([None]).map(str).to(list)
    [""]
    >>> pnq.query([(1, 2)]).map(lambda arg1, arg2: arg1, "*").to(list)
    [1]
    >>> pnq.query([{"arg1": 1, "arg2": 2}]).map(lambda arg1, arg2: arg1, "\*\*").to(list)
    [1]
    >>> pnq.query([pnq.Arguments(1, 2, name="test", age=20)]).map(lambda arg1, arg2, name, age: name, "\*\*\*").to(list)
    ["test"]
    ```
    """


def __map_nullable(self, selector):
    if selector is None:
        return self
    else:
        return map(selector, self)


def itertools__():
    from itertools import (
        chain,
        combinations,
        combinations_with_replacement,
        count,
        cycle,
        dropwhile,
        groupby,
        islice,
        permutations,
        product,
        repeat,
        starmap,
        takewhile,
        tee,
        zip_fillvalue,
        zip_longest,
    )


@mark
def select_single_node(
    self,
    node_selector=lambda x: x.node,
    return_selector=lambda nest, parent, child: (nest, parent, child),
):
    pass


@mark
def gather(self):
    """シーケンスの要素から結果を取り出し、新しいフォームに射影します。
    要素は、`concurrent.futures.Future`、または、`awaitable`である必要があります。

    `pnq.query`は、それらのインターフェースを実装しているため、本メソッドで実体化可能です。

    Usage:
    ```
    from concurretn.futures import Future
    import asyncio

    def main1():
        future = Future()
        future.set_result(0)
        return pnq.query([future]).gather().result()

    async def main2():
        async def heavy_task():
            return 1

        async def aiter():
            yield 3
            yield 4

        future = Future()
        future.set_result(0)

        tasks = pnq.query([
            future,
            heavy_task(),
            asyncio.create_task(heavy_task()),
            pnq.([1, 2]).map(lambda x: x * 2),
            pnq.(aiter()).map(lambda x: x * 2)
        ])
        return await tasks.gather()

    main()  # => [0]
    asyncio.run(main2())  # => [[0], [1], [1], [2, 4], [6, 8]]
    ```
    """


@mark
def select(self, field, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムを選択し新しいフォームに射影します。
    複数のアイテムを選択した場合は、タプルとして射影します。

    Args:

    * self: 変換対象のシーケンス
    * field: 各要素から選択するアイテム
    * fields: 各要素から選択する追加のアイテム
    * attr: 要素の属性から取得する場合はTrue

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select(0).to(list)
    [1]
    >>> pnq.query([{"id": 1, "name": "a"}]).select("id", "name").to(list)
    [(1, "a")]
    >>> pnq.query([user]).select("id", "name", attr=True).to(list)
    [(1, "a")]
    ```
    """


@mark
def select_as_tuple(self, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムまたは属性を選択し辞書として新しいフォームに射影します。
    selectと似ていますが、選択した値が１つでも必ずタプルを返します。

    Args:

    * self: 変換対象のシーケンス
    * fields: 選択するアイテムまたは属性
    * attr: 属性から取得する場合はTrueとする

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select_as_tuple(0).to(list)
    [(1,)]
    >>> pnq.query([user]).select_as_tuple("id", "name", attr=True).to(list)
    [("1", "a")]
    ```
    """


@mark
def select_as_dict(self, *fields, attr: bool = False):
    """シーケンスの各要素からアイテムまたは属性を選択し辞書として新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * fields: 選択するアイテムまたは属性
    * attr: 属性から取得する場合はTrueとする

    Usage:
    ```
    >>> pnq.query([(1, 2)]).select_as_dict(0).to(list)
    [{0: 1}]
    >>> pnq.query([user]).select_as_dict("id", "name", attr=True).to(list)
    [{"id": 1, "name": "b"}]
    ```
    """


@mark
def reflect(self, mapping, *, default=NoReturn, attr: bool = False):
    """シーケンスの各要素のフィールドを与えたマッピングに基づき辞書として新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * mapping: 元の要素のフィールドと射影先のフィールドの対応表
    * default: フィールドを取得できない場合のデフォルト値
    * attr: 属性から取得する場合はTrueとする

    Usage:
    ```
    >>> person = {"id":1, "name": "山田", "kana": "やまだ", "note": "hoge"}
    >>> pnq.query([person]).reflect({
    >>>   "id": "id",
    >>>   "name": {"name", "searchable"},
    >>>   "kana": {"kana", "searchable"},
    >>>   "note": "searchable"
    >>> }).to(list)
    >>> [{"id": 1, "name": "山田", "kana": "やまだ", "searchable": ["山田", "やまだ", "hoge"]}]
    ```
    """


@mark
def flat(self, selector=None):
    """シーケンスの各要素をイテラブルに射影し、その結果を１つのシーケンスに平坦化します。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素から平坦化する要素を選択する関数

    Usage:
    ```
    >>> pnq.query(["abc", "def"]).flat().to(list)
    >>> ["a", "b", "c", "d", "e", "f"]
    >>> countries = [{"country": "japan", "state": ["tokyo", "osaka"]}, {"country": "america", "state": ["new york", "florida"]}]
    >>> pnq.query(countries).flat(lambda x: x["state"]).to(list)
    >>> ["tokyo", "osaka", "new york", "florida"]
    ```
    """


select_many = flat
flat_map = flat


@mark
def traverse(self, selector):
    """シーケンスの各要素から再帰的に複数ノードを選択し、選択されたノードを１つのシーケンスに平坦化します。
    各ルート要素から浅い順に列挙されます。

    Args:

    * self: 変換対象のシーケンス
    * selector: 各要素から平坦化する要素を再帰的に選択する関数（戻り値はリスト等に含めて返す必要があります）

    Usage:
    ```
    >>> pnq.query(
    >>>     {"name": "a", "nodes": [{"name": "b", nodes: [{"name": c, "nodes": []}, {"name": "d", "nodes": []}}}]}]}
    >>> ).traverse(lambdax x: x["nodes"]).select("name").to(list)
    >>> ["a", "b", "c", "d"]
    ```
    """


def pivot_unstack(self, default=None):
    """行方向に並んでいるデータを列方向に入れ替える

    Args:

    * self: 変換対象のシーケンス
    * default: フィールドが存在しない場合のデフォルト値

    Usage:
    ```
    data = [
        {"name": "test1", "age": 20},
        {"name": "test2", "age": 25},
        {"name": "test3", "age": 30, "sex": "male"},
    ]
    {'name': ['test1', 'test2', 'test3'], 'age': [20, 25, 30], 'sex': [None, None, 'male']}
    ```
    """
    ...


def pivot_stack(self):
    """列方向に並んでいるデータを行方向に入れ替える

    Args:

    * self: 変換対象のシーケンス

    Usage:
    ```
    {'name': ['test1', 'test2', 'test3'], 'age': [20, 25, 30], 'sex': [None, None, 'male']}
    data = [
        {"name": "test1", "age": 20, "sex": None},
        {"name": "test2", "age": 25, "sex": None},
        {"name": "test3", "age": 30, "sex": "male"},
    ]
    ```
    """


@mark
def cast(self, type):
    """シーケンスの型注釈を変更します。この関数はエディタの型解釈を助けるためだけに存在し、何も処理を行いません。

    実際に型を変更する場合は、`map`を使用してください。

    Args:

    * self: 変換対象のシーケンス
    * type: 新しい型注釈

    Usage:
    ```
    >>> pnq.query([1]).cast(float)
    ```
    """


@mark
def enumerate(self, start: int = 0, step: int = 1):
    """シーケンスの各要素とインデックスを新しいフォームに射影します。

    Args:

    * self: 変換対象のシーケンス
    * start: 開始インデックス
    * step: 増分

    Usage:
    ```
    >>> pnq.query([1, 2]).enumerate().to(list)
    [(0, 1), (1, 2)]
    >>> pnq.query([1, 2]).enumerate(5).to(list)
    [(5, 1), (6, 2)]
    >>> pnq.query([1, 2]).enumerate(0, 10)).to(list)
    [(0, 1), (10, 2)]
    ```
    """


@mark
def group_by(self, selector=lambda x: x):
    """シーケンスの各要素からセレクタ関数でキーとバリューを取得し、キーでグループ化されたシーケンスを生成します。
    セレクタ関数を指定しない場合、各要素がすでにキーバリューのタプルであることを期待し、キーでグループ化します。

    Args:

    * self: 変換対象のシーケンス
    * selector: キーとバリューを選択する関数

    Usage:
    ```
    >>> data = [
    >>>   {"name": "banana", "color": "yellow", "count": 3},
    >>>   {"name": "apple", "color": "red", "count": 2},
    >>>   {"name": "strawberry", "color": "red", "count": 5},
    >>> ]
    >>> pnq.query(data).group_by(lambda x: x["color"], x["name"]).to(list)
    [("yellow", ["banana"]), ("red", ["apple", "strawberry"])]
    >>> pnq.query(data).select("color", "count").group_by().to(dict)
    {"yellow": [3], "red": [2, 5]}
    ```
    """


@mark
def chunk(self, size: int):
    """
    シーケンスを指定したサイズ毎に分割する。
    """


@mark
def tee(self, size: int):
    """
    ひとつのイテレータから独立したN個のイテレータを返す
    """


@mark
def join(self, right, on, select):
    pass


@mark
def group_join(self, right, on, select):
    pass


@mark
def parallel(self, func, executor=None, *, unpack="", chunksize=1):
    """シーケンスの要素を任意の関数で並列処理し、新しいフォームに射影します。
    例外が発生した場合、後続の要素はスケジューリングされません。
    その場合、実行中の処理とキューに積まれた処理の後処理は、エクゼキュータの挙動に依存します。
    詳しくは並列処理を参照ください。

    Args:

    * func: 実行する任意の関数
    * executor: 処理を実行するエクゼキュータ
    * unpack: 引数をどのように展開するか指定する
    * chunksize: cpuバウンドなエクゼキュータで有効です。指定したサイズの要素を一括処理します。

    """


@mark
def request(
    self,
    func,
    executor=None,
    *,
    unpack="",
    chunksize=1,
    # retry: int = None,
    # timeout: float = None
):
    """シーケンスの要素を任意の関数で並列処理し、新しいフォームに射影します。
    戻り値または例外は、`asyncio.Future`互換の`Response`オブジェクトに格納されます。
    詳しくは並列処理を参照ください。

    Args:

    * func: 実行する任意の関数
    * executor: 処理を実行するエクゼキュータ
    * unpack: 引数をどのように展開するか指定する
    * chunksize: cpuバウンドなエクゼキュータで有効です。指定したサイズの要素を一括処理します。

    Usage:
    ```
    >>> def do_something(id, val):
    >>>   if val:
    >>>     return 1
    >>>   else:
    >>>     raise ValueError(val)
    >>>
    >>> for res in pnq.query([{"id": 1, "val": True}, {"id": 2, "val": False}]).request(do_something):
    >>>   if res.err:
    >>>     print(f"ERROR: {res.to(dict)}")
    >>>   else:
    >>>     print(f"SUCCESS: {res.to(dict)}")
    ```
    """


@mark
def dispatch(source, func, executor=None, *, unpack="", chunksize=1, on_complete=None):
    """シーケンスの要素を任意の関数で並列処理します。
    スケジューリングされた処理はバックグラウンドで実行され、処理結果は`on_complete`に指定した関数で受け取れます。
    アプリケーションが終了する時、バックグランドの残処理がどのように扱われるかはエクゼキュータの挙動に依存します。
    `on_complete`が確実に呼び出される保証はありません。
    詳しくは並列処理を参照ください。

    Args:

    * func: 実行する任意の関数
    * executor: 処理を実行するエクゼキュータ
    * unpack: 引数をどのように展開するか指定する
    * chunksize: cpuバウンドなエクゼキュータで有効です。指定したサイズの要素を一括処理します。
    * on_complete(x): 処理結果を受け取る関数を指定します

    """


def debug(self, breakpoint=lambda x: x, printer=print):
    for v in self:
        printer(v)
        breakpoint(v)
        yield v


###########################################
# Expander set operation
###########################################

# https://docs.aws.amazon.com/ja_jp/redshift/latest/dg/r_UNION.html


@mark
def union_all(self):
    """全ての行を集合に含む"""
    pass


extend = union_all


@mark
def union(self, *iterables):
    """全ての行を集合に含み重複は許可しない"""
    pass


@mark
def union_intersect(self):
    """共通部分のみ抽出"""
    pass


@mark
def union_minus(self):
    """1つ目の問い合わせには存在するが、2つ目の問い合わせには存在しないデータを抽出
    exceptと同じ意味
    """
    pass


# difference
# symmetric_difference

# < <= > >= inclusion


def zip(self, *iterables):
    raise NotImplementedError()


def compress(self, *iterables):
    """未実装"""
    import itertools

    return itertools.compress(self, *iterables)


@mark
def cartesian(self, *iterables):
    """複数のシーケンスのデカルト積を返します。

    Args:

    * self, iterables: デカルト積を求める複数のシーケンス
    """


"""
# product（デカルト積）
from itertools import product
mark=['♠', '♥', '♦', '♣']
suu = ['A', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'J', 'Q', 'K']
tramp = [m for m in product(mark,suu)]
print(tramp)
# [('♠', 'A'), ('♠', '2'), ...]
"""


###########################################
# filtering
###########################################
def filter(self, predicate):
    """述語に基づいてシーケンスの要素をフィルタ処理します。

    Args:

    * self: フィルタ対象のシーケンス
    * predicate: 条件を満たすか検証する関数

    Usage:
    ```
    >>> pnq.query([1, 2]).filter(lambda x: x == 1).to(list)
    [1]
    >>> pnq.query({1: True, 2: False, 3: True}).filter(lambda x: x[1] == True).to(list)
    [(1, True), (3, True)]
    ```
    """


@mark
def must(self, predicate, msg=""):
    """述語に基づいてシーケンスの要素を検証します。
    検証に失敗した場合、即時に例外が発生します。

    Args:

    * self: フィルタ対象のシーケンス
    * predicate: 条件を満たすか検証する関数

    Usage:
    ```
    >>> pnq.query([1, 2]).must(lambda x: x == 1).to(list)
    raise ValueError("2")
    >>> pnq.query({1: True, 2: False, 3: True}).must(lambda x: x[1] == True).to(list)
    raise ValueError("(2, False)")
    ```
    """


@mark
def filter_type(self, *types):
    """指定した型に一致するシーケンスの要素をフィルタ処理します。
    型は複数指定することができ、`isinstance`の挙動に準じます。

    Args:

    * self: フィルタ対象のシーケンス
    * types: フィルタする型

    Usage:
    ```
    >>> pnq.query([1, False, "a"]).filter_type(int).to(list)
    [1, False]
    >>> pnq.query([1, False, "a"]).filter_type(str, bool).to(list)
    [False, "a"]
    ```
    """


@mark
def must_type(self, types):
    """シーケンスの要素が指定した型のいずれかであるか検証します。
    検証に失敗した場合、即時に例外が発生します。
    型は複数指定することができ、`isinstance`の挙動に準じます。

    Args:

    * self: フィルタ対象のシーケンス
    * types: フィルタする型

    Usage:
    ```
    >>> pnq.query([1, 2]).must_type(str, int).to(list)
    raise ValueError("1 is not str")
    ```
    """


@mark
def filter_keys(self, *keys):
    """シーケンスの要素から指定したキーの要素のみフィルタ処理します。
    このメソッドは、`list` `dict` `set`などをクエリ化した直後のみ利用できます。

    * `list`、`tuple`の場合インデックスでフィルタされ、値を返します。
    * `dict`の場合キーでフィルタされ、キーと要素を返します。
    * `set`の場合は、キーでフィルタされ、キーを返します。

    Args:

    * self: フィルタ対象のシーケンス
    * keys: フィルタするキー

    Usage:
    ```
    >>> pnq.query([1, 2]).filter_keys(1).to(list)
    [2]
    >>> pnq.query({"a": 1, "b": 2}).filter_keys("b").to(list)
    [("b", 2)]
    >>> pnq.query({"a", "b"}).filter_keys("b").to(list)
    ["b"]
    ```
    """


@mark
def must_keys(self, *keys):
    """`filter_keys`を実行し、全てのキーを取得できなかった場合例外を発生させます。
    検証が完了するまで、ストリームは保留されます。
    """


@mark
def filter_unique(self, selector=None):
    """シーケンスの要素から重複する要素を除去する。
    セレクタによって選択された値に対して重複が検証され、その値を返す。

    Args:

    * self: フィルタ対象のシーケンス
    * selector: 重複を検証する値（複数の値を検証する場合はタプル）

    Usage:
    ```
    >>> pnq.query([1, 2, 1]).filter_unique().to(list)
    [1, 2]
    >>> pnq.query([(0 , 0 , 0), (0 , 1 , 1), (0 , 0 , 2)]).unique(lambda x: (x[0], x[1])).to(list)
    [(0, 0), (0, 1)]
    ```
    """


distinct = filter_unique


@mark
def must_unique(self, selector=None):
    """シーケンスの要素から値を選択し、選択した値が重複していないか検証します。

    Args:

    * self: フィルタ対象のシーケンス
    * selector: 検証する値を選択する関数
    * immediate: 即時に例外を発生させる

    Usage:
    ```
    >>> pnq.query([1, 2, 1]).must_unique().to(list)
    raise DuplicateError("1")
    ```

    """


###########################################
# partitioning
###########################################


@mark
def take(self, count_or_range: int):
    """シーケンスから指定した範囲の要素を返します。

    Args:

    * self: 取得対象のシーケンス
    * count_or_range: シーケンスの先頭から取得する要素数または取得する範囲

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).take(2).to(list)
    [1, 2]
    >>> pnq.query([1, 2, 3]).take(range(1, 2)).to(list)
    [2]
    ```
    """


@mark
def take_while(self, predicate):
    """シーケンスの先頭から、条件の検証に失敗するまでの要素を返します。
    **検証に失敗した要素は破棄されるため、値を消費するイテレータがソースの場合は注意が必要です。**

    Args:

    * self: バイパス対象のシーケンス
    * predicate: 条件を検証する関数

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).enumerate().take_while(lambda v: v[0] < 2).select(1).to(list)
    [1, 2]
    ```
    """


@mark
def skip(self, count_or_range: int):
    """シーケンスから指定した範囲の要素をバイパスします。

    Args:

    * self: バイパス対象のシーケンス
    * count_or_range: シーケンスの先頭からバイパスする要素数またはバイパスする範囲

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).skip(1).to(list)
    [2, 3]
    >>> pnq.query([1, 2, 3]).skip(range(1, 2)).to(list)
    [1, 3]
    ```
    """


@mark
def skip_while(self, predicate):
    """シーケンスの先頭から、条件の検証に失敗するまでの要素をバイパスし、残りの要素を返します。

    Args:

    * self: バイパス対象のシーケンス
    * predicate: 条件を検証する関数

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).enumerate().skip_while(lambda v: v[0] < 1).select(1).to(list)
    [2, 3]
    ```
    """


@mark
def take_page(self, page: int, size: int):
    """シーケンスから指定した範囲の要素を返します。
    範囲はページサイズと取得対象のページから求められます。

    Args:

    * self: バイパス対象のシーケンス
    * page: 取得対象のページ（1始まり）
    * size: １ページあたりの要素数

    Usage:
    ```
    >>> pnq.query([0, 1, 2, 3, 4, 5]).take_page(page=1, size=2).to(list)
    [0, 1]
    >>> pnq.query([0, 1, 2, 3, 4, 5]).take_page(page=2, size=3).to(list)
    [3, 4, 5]
    ```
    """


###########################################
# order
###########################################


@mark
def order_by(self, key_selector=None, desc: bool = False):
    """シーケンスの要素を昇順でソートします。

    Args:

    * self: ソート対象のシーケンス
    * selector: 要素からキーを抽出する関数。複数のキーを評価する場合は、タプルを返してください。
    * desc: 降順で並べる場合はTrue

    Usage:
    ```
    >>> pnq.query([3, 2, 1]]).order_by().to(list)
    [1, 2, 3]
    >>> pnq.query([1, 2, 3]).order_by(lambda x: -x).to(list)
    [3, 2, 1]
    >>> pnq.query([1, 2, 3]).order_by(desc=True).to(list)
    [3, 2, 1]
    >>> pnq.query([(1, 2)), (2, 2), (2, 1)]).order_by(lambda x: (x[0], x[1])).to(list)
    [(1, 2)), (2, 1), (2, 2)]
    ```
    """


@mark
def order_by_fields(self, *fields, desc: bool = False, attr: bool = False):
    raise NotImplementedError()


@mark
def order_by_reverse(self):
    """シーケンスの要素を逆順にします。

    Args:

    * self: ソート対象のシーケンス

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).order_by_reverse().to(list)
    [3, 2, 1]
    ```
    """


@mark
def order_by_shuffle(self):
    """シーケンスの要素をランダムにソートします。

    Args:

    * self: ソート対象のシーケンス

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).order_by_shuffle().to(list)
    [1, 3, 2]
    >>> pnq.query([1, 2, 3]).order_by_shuffle().to(list)
    [3, 1, 2]
    ```
    """


###########################################
# statistics
###########################################
def len(self):
    """シーケンスの要素数を返します。

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).len()
    3
    ```
    """


def exists(self):
    """シーケンス内の要素の有無を確認します。

    Usage:
    ```
    >>> pnq.query([]).exists()
    False
    >>> pnq.query([1]).exists()
    True
    ```
    """


def all(self, selector=lambda x: x):
    """シーケンスの全ての要素がTrueと判定できるか評価します。要素がない場合はTrueを返します。

    * selector: 要素から検証する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).all()
    True
    >>> pnq.query([0]).all()
    False
    >>> pnq.query([1]).all()
    True
    >>> pnq.query([1, 0]).all()
    False
    >>> pnq.query([1, 2]).all()
    True
    ```
    """


def any(self, selector=lambda x: x):
    """シーケンスのいずれかの要素がTrueと判定できるか評価します。要素がない場合はFalseを返します。

    * selector: 要素から検証する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).any()
    False
    >>> pnq.query([0]).any()
    False
    >>> pnq.query([1]).any()
    True
    >>> pnq.query([1, 0]).any()
    True
    ```
    """


def contains(self, value, selector=lambda x: x) -> bool:
    """既定の等値比較子を使用して、指定した要素がシーケンスに含まれているか評価します。
    辞書をソースとした場合は、キーバリューのタプルを比較します。


    * value: 検索対象の値
    * selector: 要素から検索する値を抽出する関数

    Usage:
    ```
    >>> fruits = pnq.query(["apple", "orange"])
    >>> fruits.contains("banana")
    False
    >>> fruits.contains("apple")
    True
    >>> fruits.contains("orange")
    True
    >>> pnq.query({"a": 1, "b": 2}).contains("a")
    False
    >>> pnq.query({"a": 1, "b": 2}).contains(("a", 1))
    True
    ```
    """


def min(self, selector=lambda x: x, default=NoReturn):
    """シーケンスの要素から最小の値を取得します。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([1, 2]).min()
    1
    >>> pnq.query([]).min()
    ValueError: min() arg is an empty sequence
    >>> pnq.query([]).min(default=0)
    0
    ```
    """


def max(self, selector=lambda x: x, default=NoReturn):
    """シーケンスの要素から最大の値を取得します。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([1, 2]).max()
    2
    >>> pnq.query([]).max()
    ValueError: max() arg is an empty sequence
    >>> pnq.query([]).max(default=0)
    0
    ```
    """


def sum(self, selector=lambda x: x):
    """シーケンスの要素を合計します。

    * selector: 要素から計算する値を抽出する関数

    Usage:
    ```
    >>> pnq.query([]).sum()
    0
    >>> pnq.query([1, 2]).sum()
    3
    ```
    """


def average(
    self, selector=lambda x: x, exp: float = 0.00001, round: TH_ROUND = "ROUND_HALF_UP"
) -> Union[float, Decimal]:
    """シーケンスの要素の平均を求めます。

    Args:

    * selector: 要素から計算する値を抽出する関数
    * exp: 丸める小数点以下の桁数
    * round: 丸め方式

    Usage:
    ```
    >>> pnq.query([]).average()
    0
    >>> pnq.query([1, 2]).average()
    1.5
    ```
    """


def reduce(
    self,
    seed: T,
    op="+=",
    selector=lambda_empty,
) -> T:
    """シーケンスの要素を指定した代入演算子でシードに合成し、合成結果を返す。

    Args:

    * seed: 合成対象とする初期値(左辺)
    * op: 代入演算子または２項演算関数
    * selector: 要素から結合する値を抽出する関数（右辺）


    Usage:
    ```
    >>> pnq.query([1]).reduce(10, "+=")
    11
    >>> pnq.query([[1, 2, 3], [4, 5, 6]]).reduce([], "+=")
    [1, 2, 3, 4, 5, 6]
    >>> pnq.query([{"a": 1}, {"b": 2}]).reduce({}, "|=") # python3.9~
    {"a": 1, "b": 2}
    >>> pnq.query([1, 2, 3, 4, 5]).reduce(0, "+=", lambda x: x * 10)
    150
    >>> pnq.query([1, 2, 3, 4, 5]).reduce(0, lambda l, r: l + r, lambda x: x * 10)
    150
    ```
    """


def concat(self, selector=lambda x: x, delimiter: str = ""):
    """シーケンスの要素を文字列として連結します。
    Noneは空文字として扱われます。

    Args:

    * selector: 要素から結合する値を抽出する関数
    * delimiter: 区切り文字

    Usage:
    ```
    >>> pnq.query([]).concat()
    ""
    >>> pnq.query([1, 2]).concat()
    "12"
    >>> pnq.query(["a", "b"]).concat()
    "ab"
    >>> pnq.query(["a", None]).concat()
    "a"
    >>> pnq.query(["a", "b"]).concat(delimiter=",")
    "a,b"
    ```
    """


###########################################
# finalizer
###########################################


@mark
def result(self):
    """ストリームを評価し、結果をリストとして保存します。
    返されたリストは、クエリメソッドが実装されたリストの拡張クラスです。

    Args:

    * self: 評価するシーケンス

    Returns: `pnq.list`

    Usage:
    ```
    >>> saved = pnq.query([1, 2, 3]).result()
    [1, 2, 3]
    >>> saved.map(lambda x: x * 2).result()
    [2, 4, 6]
    ```
    """


@mark
def save(self):
    """`result`のエイリアスです。"""


@mark
def to(self, finalizer):
    """ストリームをファイナライザによって処理します。

    Args:

    * self: 評価するシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> pnq.query([1, 2, 3]).to(list)
    [1, 2, 3]
    >>> pnq.query({1: "a", 2: "b"}).to(dict)
    {1: "a", 2: "b"}
    ```
    """


async def to_async(self, finalizer):
    if self.is_async:
        return await finalizer(self)
    else:
        return await finalizer(self)


@mark
def to_task(self):
    """実体化を開始し、任意のタイミングで結果を受け取れるasyncio.Taskを返します。
    このメソッドは、非同期イテレータや並列処理などを非同期で進行しながら、
    他の処理を進める場合に役に立ちます。

    taskはawaitすることで、結果を受け取ることができます。

    Usage:
    ```
    >>> task = pnq.query(aiter).to_task()
    >>> await do_something()
    >>> await task
    [1, 2, 3]
    ```
    """


@mark
def to_lifespan(self):
    """イテレータを生存期間つきのイテレータ（コンテキスト）にします。
    このメソッドは、無限イテレータの開始・停止の管理に役立ちます。

    callbackが設定されている場合は、ルートイテレータが停止されるまで生存し、
    callbackが指定されていない場合は、直ちにイテレートを停止します。

    >>> queue = pnq.Queue()
    >>> async with pnq.query(queue).to_lifespan(print, callback=queue.stop):
    >>>   await queue.put(1)
    >>>   await queue.put(2)
    """


@mark
def lazy(self, finalizer):
    """ファイナライザの実行するレイジーオブジェクトを返します。
    レイジーオブジェクトをコールすると同期実行され、`await`すると非同期実行されます。

    Args:

    * self: バイパス対象のシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> lazy = pnq.query([1, 2, 3]).lazy(list)
    >>> lazy()
    [1, 2, 3]
    >>> await lazy
    [1, 2, 3]
    >>> lazy = pnq.query([1, 2, 3]).lazy(pnq.actions.first)
    >>> await lazy
    1
    ```
    """
    return finalizer(x for x in self)


async def to_async(self, cls):
    """非同期ストリームをファイナライザによって処理します。
    クエリが非同期処理を要求する場合のみ使用してください。

    Args:

    * self: バイパス対象のシーケンス
    * finalizer: イテレータを受け取るクラス・関数

    Returns: ファイナライザが返す結果

    Usage:
    ```
    >>> await pnq.query([1, 2, 3]).to_async(list)
    [1, 2]
    >>> await pnq.query({1: "a", 2: "b"}).to_async(dict)
    {1: "a", 2: "b"}
    ```
    """


def each(self, func=lambda x: x, unpack=""):
    """シーケンスの各要素を指定した関数で逐次的に処理します。
    `for x in iterable: ...`のショートカットとして機能します。
    関数を指定しない場合、単にイテレーションを実行します。

    非同期関数を実行するには、クエリを非同期化する必要があります。

    Args:

    * func: 要素を処理する関数
    * unpack: 引数をどのように展開するか指定する

    Returns: `None`

    Usage:
    ```
    >>> pnq.query([1,2]).each()
    >>> pnq.query([1,2]).each(print)
    1
    2
    >>> await pnq.query([1,2]).\_.each(sync_func)
    >>> await pnq.query([1,2]).\_.each(async_func)
    ```
    """


@mark
def get(self, key, default=NoReturn):
    """リストや辞書などの`__getitem__`を呼び出します。セットでも使用でき、キーが存在する場合そのままキーを返します。
    デフォルトを設定した場合、キーが存在しない場合はデフォルトを返します。

    このメソッドは実体化している状態でのみ使用することができ、クエリ化されている状態では使用できません。

    Args:

    * key: キー
    * default: キーが存在しない場合に返すデフォルト値

    Usage:
    ```
    >>> data = pnq.query({"a", "b", "c"})
    >>> data.get("a")
    "a"
    >>> data.get("d")
    raise KeyError("d")
    >>> data.get("d", 10)
    10
    ```
    """
    try:
        return self[key]
    except (IndexError, KeyError, NotFoundError):
        pass

    except Exception:
        if isinstance(self, set) and key in self:
            return key

    if default is NoReturn:
        raise NotFoundError(key)
    else:
        return default


@mark
def one(self):
    """シーケンス内の要素が１つであることを検証し、その要素を返します。
    検証に失敗した場合は、例外が発生します。
    デフォルト値を設定した場合は、要素が存在しない場合にデフォルト値を返します。

    `one`関数は、１つの要素であるか検証するために２つ目の要素を取り出そうとします。
    ソースとなるイテラブルが値を消費する実装だと、２つの要素が失われる可能性があることに注意してください。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).one()
    raise NoElementError("...")
    >>> pnq.query([1]).one()
    1
    >>> pnq.query([1, 2]).one()
    raise NotOneElementError("...")
    >>> pnq.query([]).one(None)
    None
    >>> pnq.query([1, 2]).one(None)
    raise NotOneElementError("...")
    ```
    """


@mark
def first(self):
    """シーケンス内の最初の要素を返します。
    要素が存在しない場合は、例外が発生します。

    セットをソースとした場合、セットは順序を保持しないため、順序性は期待できません。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).first()
    raise NoElementError("...")
    >>> pnq.query([1]).first()
    1
    >>> pnq.query([1, 2]).first()
    1
    >>> pnq.query([]).first(None)
    None
    ```
    """


@mark
def last(self):
    """シーケンス内の最後の要素を返します。
    要素が存在しない場合は、例外が発生します。

    セットをソースとした場合、セットは順序を保持しないため、順序性は期待できません。

    Args:

    * default: 要素が存在しない場合に返す値

    Usage:
    ```
    >>> pnq.query([]).last()
    raise NoElementError("...")
    >>> pnq.query([1]).last()
    1
    >>> pnq.query([1, 2]).last()
    2
    >>> pnq.query([]).last(None)
    None
    ```
    """


@mark
def get_or(self, key, default):
    try:
        return get(self, key)
    except NotFoundError:
        return default


@mark
def one_or(self, default):
    try:
        return one(self)
    except NoElementError:
        return default


@mark
def first_or(self, default):
    try:
        return first(self)
    except NoElementError:
        return default


@mark
def last_or(self, default):
    try:
        return last(self)
    except NoElementError:
        return default


@mark
def get_or_raise(self, key, exc: Union[str, Exception]):
    """基本的な動作は`get`を参照ください。
    KeyErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: KeyError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).get_or_raise(0, Exception(f"Not Exist Key: 0"))
    raise Exception("Not Exist Key: 0")
    ```
    """


@mark
def one_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`one`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。
    NotOneElementErrorはキャッチしません。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).one_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """


@mark
def first_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`first`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).first_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """


@mark
def last_or_raise(self, exc: Union[str, Exception]):
    """基本的な動作は`last`を参照ください。
    NoElementErrorが発生した時、任意の例外を発生させます。

    Args:

    * exc: NoElementError時に発生させる例外

    Usage:
    ```
    >>> pnq.query([]).last_or_raise(0, Exception("No exists."))
    raise Exception("No exists.")
    ```
    """


###########################################
# sleeping
###########################################
def sleep(self, seconds: float):
    from time import sleep as sleep_


async def sleep_async(self, seconds: float):
    from asyncio import sleep as sleep_



import asyncio
from typing import Iterator, TypeVar

T = TypeVar("T")


class QuerySyncToAsync(Query[T]):
    """同期イテレータを非同期イテレータに変換します。
    もしくは、同期イテレータを取得できない場合、非同期イテレータの取得を試みます。"""

    iter_type = IterType.ASYNC

    def __init__(self, source: Iterable[T]):
        # super().__init__(self)
        self.source = source
        self.run_iter_type = self.iter_type

    def _impl_iter(self):
        raise NotImplementedError()

    def _impl_aiter(self):
        it = None
        try:
            it = iter(self.source)
        except Exception:
            pass

        if it:
            return sync_to_async_iterator(it)
        else:
            return self.source.__aiter__()


class ToSync:
    """非同期イテレータを同期イテレータに変換する"""

    iter_type = IterType.SYNC

    def __iter__(self) -> Iterator[T]:
        if self.loop:
            loop = self.loop
        else:
            loop = asyncio.new_event_loop()  # +python3.8

        try:
            aiter = self.source.__aiter__()
        except Exception:
            loop.close()

        async def get_next(aiter):
            try:
                obj = await aiter.__anext__()
                return False, obj
            except StopAsyncIteration:
                return True, None

        try:
            while True:
                done, obj = loop.run_until_complete(get_next(aiter))
                if done:
                    break
                yield obj
        finally:
            loop.close()



from typing import (
    TYPE_CHECKING,
    AbstractSet,
    AsyncIterable,
    Callable,
    Iterable,
    List,
    Mapping,
    NoReturn,
    Sequence,
    TypeVar,
    Union,
)

from pnq import selectors
from pnq._itertools.common import name_as
from pnq.exceptions import NotFoundError
from pnq.types import Arguments

from . import _async as A
from . import _sync as S
from .core import IterType
from .core import Query as QueryBase
from .core import QueryDict, QuerySeq, QuerySet


def no_implement(*args, **kwargs):
    raise NotImplementedError()


F = TypeVar("F", bound=Callable)


class Staticmethod:
    def __or__(self, other: F) -> F:
        return staticmethod(other)  # type: ignore


exports = []


def export(cls):
    exports.append(cls)
    return cls


T = TypeVar("T")
sm = Staticmethod()


#################################
# query
#################################
class Query(QueryBase[T]):
    _iter_type = IterType.BOTH

    if TYPE_CHECKING:
        _sit: Callable
        _ait: Callable

    def __init__(self, source: Union[Iterable[T], AsyncIterable[T]]):
        super().__init__(source)

    _args = Arguments()

    def _impl_iter(self):
        x = self._args
        return self._sit(self.source, *x.args, **x.kwargs)

    def _impl_aiter(self):
        x = self._args
        return self._ait(self.source, *x.args, **x.kwargs)


@export
class Map(Query):
    _ait = sm | A.queries._map
    _sit = sm | S.queries._map

    def __init__(self, source, selector, unpack=""):
        super().__init__(source)
        self._args = Arguments(selector, unpack)


@export
class Select(Map):
    def __init__(self, source, *args, attr=False):
        selector = selectors.select(*args, attr=attr)
        super().__init__(source, selector)


@export
class SelectAsTuple(Map):
    def __init__(self, source, *args, attr=False):
        selector = selectors.select_as_tuple(*args, attr=attr)
        super().__init__(source, selector)


@export
class SelectAsDict(Map):
    def __init__(self, source, *args, attr=False, default=NoReturn):
        selector = selectors.select_as_dict(*args, attr=attr, default=default)
        super().__init__(source, selector)


@export
class Reflect(Map):
    def __init__(self, source, mapping, attr=False):
        selector = selectors.reflect(mapping, attr=attr)
        super().__init__(source, selector)


@export
class Flat(Query):
    _ait = sm | A.queries.flat
    _sit = sm | S.queries.flat

    def __init__(self, source, selector=None):
        super().__init__(source)
        self.selector = selector
        self._args = Arguments(selector)


@export
class Traverse(Query):
    _ait = sm | A.queries.traverse
    _sit = sm | S.queries.traverse

    def __init__(self, source, selector):
        super().__init__(source)
        self._args = Arguments(selector)


@export
class PivotUnstack(Query):
    _ait = sm | A.queries.pivot_unstack
    _sit = sm | S.queries.pivot_unstack

    def __init__(self, source, default=None):
        super().__init__(source)
        self._args = Arguments(default)


@export
class PivotStack(Query):
    _ait = sm | A.queries.pivot_stack
    _sit = sm | S.queries.pivot_stack


@export
class Enumerate(Query):
    _ait = sm | A.queries._enumerate
    _sit = sm | S.queries._enumerate

    def __init__(self, source, start=0, step=1):
        super().__init__(source)
        self._args = Arguments(start, step)


@export
class GroupBy(Query):
    _ait = sm | A.queries.group_by
    _sit = sm | S.queries.group_by

    def __init__(self, source, selector=None):
        super().__init__(source)
        self._args = Arguments(selector)


@export
class Join(Query):
    _ait = sm | A.queries.join
    _sit = sm | S.queries.join


@export
class InnerJoin(Query):
    _ait = sm | A.queries.inner_join
    _sit = sm | S.queries.inner_join

    def __init__(self, source, right):
        super().__init__(source)
        self._args = Arguments(right)


@export
class GroupJoin(Query):
    _ait = sm | A.queries.group_join
    _sit = sm | S.queries.group_join


@export
class Chain(Query):
    _ait = sm | A.queries.chain
    _sit = sm | S.queries.chain

    def __init__(self, source, *iterables):
        super().__init__(source)
        self._args = Arguments(*iterables)


@export
class Chunk(Query):
    _ait = sm | A.queries.chunk
    _sit = sm | S.queries.chunk

    def __init__(self, source, size: int):
        super().__init__(source)
        if size <= 0:
            raise ValueError("count must be greater than 0")

        self._args = Arguments(size)


@export
class Tee(Query):
    _ait = sm | A.queries.tee
    _sit = sm | S.queries.tee

    def __init__(self, source, size: int):
        super().__init__(source)
        if size <= 0:
            raise ValueError("count must be greater than 0")

        self._args = Arguments(size)


@export
class Debug(Query):
    _ait = sm | A.queries.debug
    _sit = sm | S.queries.debug

    def __init__(self, source, breakpoint=lambda x: x, printer=print):
        super().__init__(source)
        self._args = Arguments(breakpoint, printer)


@export
class DebugPath(Query):
    def __init__(self, source, func=lambda x: -10, async_func=lambda x: 10):
        super().__init__(source)
        self._args = Arguments(func, async_func)

    def _impl_iter(self):
        func = self._args.args[0]
        for v in self.source:
            yield func(v)

    async def _impl_aiter(self):
        async_func = self._args.args[1]
        async for v in self.source:
            yield async_func(v)


@export
class UnionAll(Query):
    _ait = sm | A.queries.union_all
    _sit = sm | S.queries.union_all


# class Extend(Query):
#     _ait = A.queries.extend
#     _sit = S.queries.extend


@export
@name_as("Union")
class _Union(Query):
    _ait = sm | A.queries.union
    _sit = sm | S.queries.union


@export
class UnionIntersect(Query):
    _ait = sm | A.queries.union_intersect
    _sit = sm | S.queries.union_intersect


@export
class UnionMinus(Query):
    _ait = sm | A.queries.union_intersect
    _sit = sm | S.queries.union_intersect


@export
class Zip(Query):
    _ait = sm | no_implement
    _sit = sm | S.queries._zip


@export
class Compress(Query):
    _ait = sm | A.queries.compress
    _sit = sm | S.queries.compress


@export
class Cartesian(Query):
    _ait = sm | A.queries.cartesian
    _sit = sm | S.queries.cartesian

    def __init__(self, source, *iterables):
        super().__init__(source)
        self._args = Arguments.from_obj(iterables, {})


@export
class Filter(Query):
    _ait = sm | A.queries._filter
    _sit = sm | S.queries._filter

    def __init__(self, source, predicate, unpack=""):
        super().__init__(source)
        self._args = Arguments(predicate, unpack)


@export
class Must(Query):
    _ait = sm | A.queries.must
    _sit = sm | S.queries.must

    def __init__(self, source, predicate, msg: str = ""):
        super().__init__(source)
        self._args = Arguments(predicate, msg)


@export
class FilterType(Query):
    _ait = sm | A.queries.filter_type
    _sit = sm | S.queries.filter_type

    def __init__(self, source, *types):
        super().__init__(source)
        types = tuple(None.__class__ if x is None else x for x in types)
        self._args = Arguments.from_obj(types, {})


@export
class MustType(Query):
    _ait = sm | A.queries.must_type
    _sit = sm | S.queries.must_type

    def __init__(self, source, *types):
        super().__init__(source)
        types = tuple(None.__class__ if x is None else x for x in types)
        self._args = Arguments.from_obj(types, {})


@export
class FilterUnique(Query):
    _ait = sm | A.queries.filter_unique
    _sit = sm | S.queries.filter_unique

    def __init__(self, source, selector=None):
        super().__init__(source)
        self._args = Arguments(selector)


@export
class MustUnique(Query):
    _ait = sm | A.queries.must_unique
    _sit = sm | S.queries.must_unique

    def __init__(self, source, selector=None):
        super().__init__(source)
        self._args = Arguments(selector)


def get_many_for_mapping(query_dict, keys):
    """"""
    undefined = object()
    for key in keys:
        obj = query_dict.get(key, undefined)
        if obj is not undefined:
            yield key, obj


def get_many_for_sequence(query_seq, keys):
    """"""
    for key in keys:
        try:
            yield query_seq[key]
        except IndexError:
            ...


def get_many_for_set(query_set, keys):
    """"""
    for key in keys:
        if key in query_set:
            yield key


@export
class FilterKeys(Query):
    _ait = sm | A.queries.filter_keys
    _sit = sm | S.queries.filter_keys

    def __init__(self, source, *keys):
        super().__init__(source)
        keys = dict.fromkeys(keys, None)  # use dict. because set has no order.
        self.keys = keys

        if isinstance(self.source, (QuerySeq, QueryDict, QuerySet)):
            source = self.source.source
        else:
            source = self.source
        if isinstance(source, Mapping):
            filter = get_many_for_mapping
        elif isinstance(source, Sequence):
            filter = get_many_for_sequence
        elif isinstance(source, AbstractSet):
            filter = get_many_for_set
        else:
            raise TypeError(f"{source} is not QuerySeq, QueryDict or QuerySet")

        self._ref = source
        self._filter = filter
        self._args = Arguments.from_obj(keys, {})

    def _impl_iter(self):
        return self._filter(self._ref, self.keys)

    async def _impl_aiter(self):
        for x in self._impl_iter():
            yield x


def get_for_dict(obj: dict, k, default):
    return obj.get(k, default)


def get_for_seq(obj: list, k, default):
    try:
        return obj[k]
    except IndexError:
        return default


def get_for_set(obj: set, k, default):
    if k in obj:
        return k
    else:
        return default


@export
class MustKeys(Query):
    _ait = sm | A.queries.must_keys
    _sit = sm | S.queries.must_keys

    def __init__(self, source, *keys, typ: str):
        # TODO: remove typ

        super().__init__(source)
        keys = dict.fromkeys(keys, None)  # use dict. because set has no order.
        self.keys = keys
        self._args = Arguments.from_obj(keys, {})

        if isinstance(self.source, (QuerySeq, QueryDict, QuerySet)):
            source = self.source.source
        else:
            source = self.source
        if isinstance(source, Mapping):
            getter = get_for_dict
        elif isinstance(source, Sequence):
            getter = get_for_seq
        elif isinstance(source, AbstractSet):
            getter = get_for_set
        else:
            raise TypeError(f"{source} is not QuerySeq, QueryDict or QuerySet")

        self._ref = source
        self._getter = getter

    def _impl_iter(self):
        source = self._ref
        not_exists = set()
        key_values = []
        undefined = object()
        getter = self._getter

        for k in self.keys:
            val = getter(source, k, undefined)
            if val is undefined:
                not_exists.add(k)
            else:
                key_values.append((k, val))

        if not_exists:
            raise NotFoundError(str(not_exists))

        if getter is get_for_dict:
            for k, v in key_values:
                yield k, v
        elif getter == get_for_seq:
            for k, v in key_values:
                yield v
        elif getter == get_for_set:
            for k, v in key_values:
                yield k
        else:
            raise TypeError(f"unknown type")

    async def _impl_aiter(self):
        for x in self._impl_iter():
            yield x


@export
class Take(Query):
    _ait = sm | A.queries.take
    _sit = sm | S.queries.take

    def __init__(self, source, count_or_range: Union[int, range]):
        super().__init__(source)
        if isinstance(count_or_range, range):
            r = range(
                max(count_or_range.start, 0),
                max(count_or_range.stop, 0),
                max(count_or_range.step, 1),
            )
        else:
            if count_or_range < 0:
                count_or_range = 0

            r = range(count_or_range)

        self._args = Arguments(r)


@export
class Skip(Take):
    _ait = sm | A.queries.skip
    _sit = sm | S.queries.skip


@export
class TakePage(Take):
    _ait = sm | A.queries.take
    _sit = sm | S.queries.take

    def __init__(self, source, page: int, size: int):
        if page < 1:
            raise ValueError("page must be >= 1")
        if size < 0:
            raise ValueError("size must be >= 0")
        start = (page - 1) * size
        stop = start + size
        super().__init__(source, range(start, stop))


@export
class TakeWhile(Query):
    _ait = sm | A.queries.take_while
    _sit = sm | S.queries.take_while

    def __init__(self, source, predicate):
        super().__init__(source)
        self._args = Arguments(predicate)


@export
class SkipWhile(TakeWhile):
    _ait = sm | A.queries.skip_while
    _sit = sm | S.queries.skip_while


@export
class OrderByMap(Query):
    _ait = sm | A.queries.order_by
    _sit = sm | S.queries.order_by

    def __init__(self, source, selector=None, desc: bool = False):
        super().__init__(source)
        self._args = Arguments(selector, desc)


@export
class OrderBy(OrderByMap):
    _ait = sm | A.queries.order_by
    _sit = sm | S.queries.order_by

    def __init__(self, source, *fields, desc: bool = False, attr: bool = False):
        if not len(fields):
            selector = None
        else:
            if attr:
                selector = selectors.select_from_attr(*fields)
            else:
                selector = selectors.select_from_item(*fields)
        super().__init__(source, selector, desc)


@export
class OrderBySelect(OrderBy):
    _ait = sm | A.queries.order_by
    _sit = sm | S.queries.order_by

    def __init__(self, source, *fields, attr: bool = False, desc: bool = False):
        if not len(fields):
            selector = None
        else:
            if attr:
                selector = selectors.select_from_attr(*fields)
            else:
                selector = selectors.select_from_item(*fields)

        super().__init__(source, selector, desc)


@export
class OrderByReverse(Query):
    _ait = sm | A.queries.order_by_reverse
    _sit = sm | S.queries.order_by_reverse


@export
class OrderByShuffle(Query):
    _ait = sm | A.queries.order_by_shuffle
    _sit = sm | S.queries.order_by_shuffle


@export
class Sleep(Query):
    _ait = sm | A.queries.sleep
    _sit = sm | S.queries.sleep

    def __init__(self, source, seconds):
        super().__init__(source)
        self._args = Arguments(seconds)


@export
class Gather(Query):
    _ait = sm | A.concurrent.gather
    _sit = sm | S.concurrent.gather


@export
class Request(Query):
    # _ait = sm | A.queries.request
    # _sit = sm | S.queries.request
    _ait = sm | A.concurrent.request
    _sit = sm | S.concurrent.request

    # def __init__(self, source, func, timeout: float = None, retry: int = 0):
    #     super().__init__(source)
    #     self._args = Arguments(func, retry)

    def __init__(
        self,
        source: AsyncIterable[T],
        func,
        executor=None,
        *,
        unpack="",
        chunksize=1,
        retry: int = None,
        timeout: float = None,
    ):
        super().__init__(source)
        self._args = Arguments(
            func,
            executor,
            unpack=unpack,
            chunksize=chunksize,
            retry=retry,
            timeout=timeout,
        )


@export
class Parallel(Query):
    """
    PEP 3148
    I/Oバウンドを効率化するにはchunksizeを1にする。
    CPUをフル活用するにはchunksizeを大きくする。
    ProcessPoolのみchunksizeは有効
    スクレイピングなどの重い処理を並列化する parallel([], func, chunksize=1)
    簡単な計算を大量に行う処理を並列化する parallel([], func, chunksize=100)
    """

    _ait = sm | A.concurrent.parallel
    _sit = sm | S.concurrent.parallel

    def __init__(self, source, func, executor=None, *, unpack="", chunksize=1):
        super().__init__(source)
        self._args = Arguments(func, executor, unpack=unpack, chunksize=chunksize)


# いらない！！！
@export
class AsyncMap(Query):
    iter_type = IterType.ASYNC

    def __init__(self, source, selector):
        super().__init__(source)
        self.selector = selector

    async def _impl_aiter(self):
        selector = self.selector
        async for x in self.source:
            yield await selector(x)


# TODO： いらない！！！
@export
class Lazy(Query):
    def __init__(self, source, finalizer, *args, **kwargs):
        super().__init__(source)
        self.finalizer = finalizer
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        return self.finalizer(self.source, *self.args, **self.kwargs)

    def __await__(self):
        coro = self.finalizer(self.source, *self.args, **self.kwargs)
        return coro.__await__()


#################################
# finalizer
#################################
class FinalizerBase:
    empty = S.finalizers.empty
    exists = S.finalizers.exists
    len = S.finalizers._len
    all = S.finalizers._all
    any = S.finalizers._any
    sum = S.finalizers._sum
    min = S.finalizers._min
    max = S.finalizers._max
    average = S.finalizers.average
    reduce = S.finalizers.reduce
    accumulate = S.finalizers.accumulate
    find = S.finalizers.find
    contains = S.finalizers.contains  # TODO: findかどっちかに統一する
    concat = S.finalizers.concat
    one = S.finalizers.one
    one_or = S.finalizers.one_or
    one_or_raise = S.finalizers.one_or_raise
    first = S.finalizers.first
    first_or = S.finalizers.first_or
    first_or_raise = S.finalizers.first_or_raise
    last = S.finalizers.last
    last_or = S.finalizers.last_or
    last_or_raise = S.finalizers.last_or_raise
    each = S.finalizers.each
    dispatch = S.concurrent.dispatch
    lazy = Lazy
    to = S.finalizers.to
    to_file = S.io.to_file
    to_csv = S.io.to_csv
    to_json = S.io.to_json
    to_jsonl = S.io.to_jsonl


class AsyncFinalizerBase:
    empty = A.finalizers.empty
    exists = A.finalizers.exists
    len = A.finalizers._len
    all = A.finalizers._all
    any = A.finalizers._any
    sum = A.finalizers._sum
    min = A.finalizers._min
    max = A.finalizers._max
    average = A.finalizers.average
    reduce = A.finalizers.reduce
    accumulate = A.finalizers.accumulate
    find = A.finalizers.find
    contains = A.finalizers.contains
    concat = A.finalizers.concat
    one = A.finalizers.one
    one_or = A.finalizers.one_or
    one_or_raise = A.finalizers.one_or_raise
    first = A.finalizers.first
    first_or = A.finalizers.first_or
    first_or_raise = A.finalizers.first_or_raise
    last = A.finalizers.last
    last_or = A.finalizers.last_or
    last_or_raise = A.finalizers.last_or_raise
    each = A.finalizers.each
    dispatch = A.concurrent.dispatch
    lazy = Lazy
    to = A.finalizers.to
    to_file = A.io.to_file
    to_csv = A.io.to_csv
    to_json = A.io.to_json
    to_jsonl = A.io.to_jsonl


class Finalizer(FinalizerBase, Iterable[T]):
    def __init__(self, source: Iterable[T]):
        self.source = source

    def __iter__(self):
        return self.source.__iter__()

    def result(self, timeout: float = None) -> List[T]:
        return list(self.source)


class AsyncFinalizer(AsyncFinalizerBase, AsyncIterable[T]):
    def __init__(self, source: AsyncIterable[T]):
        self.source = source

    def __aiter__(self):
        return self.source.__aiter__()

    def __await__(self):
        return self._result_async().__await__()

    async def _result_async(self):
        return [x async for x in self]



# from .base import exceptions
# from .base.requests import Response
from . import concurrent, exceptions, operators, selectors
from ._itertools.requests import Response
from .io import from_csv, from_jsonl
from .queries import PnqList as list
from .queries import query, run
from .types import Arguments, exitstack

# from . import actions



import asyncio
from concurrent.futures import Future as _Future
from typing import Iterable, Mapping

from typing_extensions import Protocol


class PArguments(Protocol):
    args: Iterable
    kwargs: Mapping


class Futurable:
    def __future__(self) -> _Future:
        ...


class AsyncFuturable:
    def __afuture__(self) -> asyncio.Future:
        ...



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pnq',
 'pnq.__base',
 'pnq._itertools',
 'pnq._itertools._async',
 'pnq._itertools._sync',
 'pnq._itertools._sync_generate',
 'pnq.aio',
 'pnq.concurrent',
 'pnq.ds',
 'pnq.operators',
 'pnq.selectors',
 'pnq.types']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'pnq',
    'version': '0.0.19',
    'description': '',
    'long_description': '# PNQ\n\n[![CI](https://github.com/sasano8/pnq/actions/workflows/test.yml/badge.svg)](https://github.com/sasano8/pnq/actions)\n[![pypi](https://img.shields.io/pypi/v/pnq.svg)](https://pypi.python.org/pypi/pnq)\n[![Downloads](https://pepy.tech/badge/pnq/month)](https://pepy.tech/project/pnq)\n\nPNQ is a Python implementation like Language Integrated Query (LINQ).\n\n!!! danger\n    PNQはベータ版です。\n\n    - 現在、ドキュメントとAPIが一致していません。\n    - ライブラリが十分な品質に到達するまで、頻繁に内部実装やAPIが更新される恐れがあります。\n    - 本番環境では利用しないでください。\n\n---\n\n\n## Features\n\n- コレクション操作に関する多彩な操作\n- アクセシブルなインタフェース\n- 型ヒントの活用\n- 非同期ストリームに対応\n\n## Similar tools\n\n- [PyFunctional](https://github.com/EntilZha/PyFunctional)\n- [linqit](https://github.com/avilum/linqit)\n- [python-linq](https://github.com/jakkes/python-linq)\n- [aioitertools](https://github.com/omnilib/aioitertools)\n- [asyncstdlib](https://github.com/maxfischer2781/asyncstdlib)\n- [asq](https://github.com/sixty-north/asq)\n\n## Documentation\n\n- See [documentation](https://sasano8.github.io/pnq/) for more details.\n\n## Dependencies\n\n- Python 3.8+\n\n## Installation\n\nInstall with pip:\n\n```shell\n$ pip install pnq\n```\n\n## Getting Started\n\n``` python\nimport pnq\n\nfor x in pnq.query([1, 2, 3]).map(lambda x: x * 2):\n    print(x)\n# => 2, 4, 6\n\npnq.query([1, 2, 3]).map(lambda x: x * 2).result()\n# => [2, 4, 6]\n\npnq.query([1, 2, 3]).filter(lambda x: x == 3).one()\n# => 2\n```\n\n``` python\nimport asyncio\nimport pnq\n\nasync def aiter():\n    yield 1\n    yield 2\n    yield 3\n\nasync def main():\n    async for x in pnq.query(aiter()).map(lambda x: x * 2):\n        print(x)\n    # => 2, 4, 6\n\n    await pnq.query(aiter()).map(lambda x: x * 2)\n    # => [2, 4, 6]\n\n    await pnq.query(aiter()).filter(lambda x: x == 3)._.one()\n    # => 3\n\nasyncio.run(main())\n```\n\n## release note\n\n### v0.0.1 (2021-xx-xx)\n\n* Initial release.\n',
    'author': 'sasano8',
    'author_email': 'y-sasahara@ys-method.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/sasano8/pnq',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import asyncio
import logging
from asyncio import iscoroutinefunction
from collections import deque
from concurrent.futures import Future
from concurrent.futures import ProcessPoolExecutor as _ProcessPoolExecutor
from concurrent.futures import ThreadPoolExecutor as _ThreadPoolExecutor
from contextlib import AsyncExitStack, ExitStack
from functools import partial
from typing import TYPE_CHECKING, Union

from . import tools
from .protocols import Executor, PExecutable, PExecutor


class OverrideExecutor:
    def __executor__(self):
        return self

    @property
    def _loop(self):
        return self._get_loop()

    _get_loop = asyncio.get_running_loop

    async def __aenter__(self):
        loop = asyncio.get_running_loop()
        self._get_loop = lambda: loop
        result = super().__enter__()
        return result

    async def __aexit__(self, *args, **kwargs):
        result = super().__exit__(*args, **kwargs)
        await asyncio.sleep(0.01)
        return result

    if TYPE_CHECKING:

        def submit(self, *args, **kwargs):
            ...

    def submit_async(self, func, *args, **kwargs):
        return self.submit(self._submit_async, func, *args, **kwargs)

    @staticmethod
    def _submit_async(func, *args, **kwargs):
        coro = func(*args, **kwargs)
        return asyncio.run(coro)

    def asubmit(self, func, *args, **kwargs):
        future = self.submit(func, *args, **kwargs)
        return asyncio.wrap_future(future, loop=self._loop)

    def asubmit_async(self, func, *args, **kwargs):
        future = self.submit_async(func, *args, **kwargs)
        return asyncio.wrap_future(future, loop=self._loop)

    def request(self, func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):
            future = self.submit_async(func, *args, **kwargs)
        else:
            future = self.submit(func, *args, **kwargs)
        return future

    def arequest(self, func, *args, **kwargs):
        if asyncio.iscoroutinefunction(func):
            future = self.asubmit_async(func, *args, **kwargs)
        else:
            future = self.asubmit(func, *args, **kwargs)
        return future


def get_executor(executable: Union[Executor, PExecutable]) -> PExecutor:
    if hasattr(executable, "__executor__"):
        return executable.__executor__()  # type: ignore

    return ExecutorWrapper(executable)


def asyncio_run(coroutine_function, *args, **kwargs):
    return asyncio.run(coroutine_function(*args, **kwargs))


class ExecutorWrapper(PExecutor, PExecutable):
    ASYNC_RUNNER = asyncio_run

    def __init__(self, executor: Executor):
        self.__post_init__(executor)

    def __post_init__(self, executor: Executor):
        self.executor = executor

        if isinstance(executor, _ProcessPoolExecutor):
            self._is_closed = lambda: executor._shutdown_thread  # type: ignore
            self._is_cpubound = True
            self._is_async_only = False
        elif isinstance(executor, _ThreadPoolExecutor):
            self._is_closed = lambda: executor._shutdown
            self._is_cpubound = False
            self._is_async_only = False
        else:
            raise TypeError()

        try:
            self._loop = asyncio.get_running_loop()
        except Exception:
            ...

    def __executor__(self) -> PExecutor:
        return self

    @property
    def running_task_count(self) -> int:
        raise NotImplementedError()

    @property
    def is_full(self) -> bool:
        raise NotImplementedError()

    @property
    def is_closed(self) -> bool:
        return self._is_closed()

    @property
    def max_workers(self) -> int:
        return self.executor._max_workers

    @property
    def is_cpubound(self) -> bool:
        return self._is_cpubound

    @property
    def is_async_only(self):
        return self._is_async_only

    def submit(self, func, *args, **kwargs):
        if iscoroutinefunction(func):
            return self.executor.submit(
                self.__class__.ASYNC_RUNNER, func, *args, **kwargs
            )
        else:
            return self.executor.submit(func, *args, **kwargs)

    def asubmit(self, func, *args, **kwargs):
        loop = self._loop
        future = self.submit(func, *args, **kwargs)
        return asyncio.wrap_future(future, loop=loop)

    def map(self, iterable, func, unpack="", timeout=None):
        return tools.map(self, iterable, func, unpack, timeout=timeout)


class PoolContext:
    POOL_FACTORY = None
    POOL_WRAPEER = lambda x: x

    def __init__(self, max_workers=None, *args, **kwargs):
        self.args = (max_workers, *args)
        self.kwargs = kwargs
        self.executor = None

    def __enter__(self) -> PExecutor:
        if self.executor:
            raise RuntimeError()

        with ExitStack() as stack:
            executor = self.POOL_FACTORY(*self.args, **self.kwargs)
            stack.enter_context(executor)
            executor_wrapper = self.POOL_WRAPEER(executor)
            self.executor = executor
            stack.pop_all()
        return executor_wrapper

    async def __aenter__(self) -> PExecutor:
        if self.executor:
            raise RuntimeError()

        async with AsyncExitStack() as stack:
            executor = self.POOL_FACTORY(*self.args, **self.kwargs)
            stack.enter_async_context(executor)
            executor_wrapper = self.POOL_WRAPEER(executor)
            self.executor = executor
            stack.pop_all()
        return executor_wrapper

    def __exit__(self, *args, **kwargs):
        return self.executor.__exit__(*args, **kwargs)

    async def __aexit__(self, *args, **kwargs):
        result = await self.executor.__aexit__(*args, **kwargs)
        await asyncio.sleep(0.01)
        return result

    @property
    def is_async_only(self):
        raise NotImplementedError()


class ProcessPoolExecutor(PoolContext):
    POOL_FACTORY = _ProcessPoolExecutor  # type: ignore
    POOL_WRAPEER = ExecutorWrapper

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, *args, **kwargs):
        result = self.__exit__(*args, **kwargs)
        await asyncio.sleep(0.01)
        return result


class ThreadPoolExecutor(PoolContext):
    POOL_FACTORY = _ThreadPoolExecutor  # type: ignore
    POOL_WRAPEER = ExecutorWrapper

    async def __aenter__(self):
        return self.__enter__()

    async def __aexit__(self, *args, **kwargs):
        result = self.__exit__(*args, **kwargs)
        await asyncio.sleep(0.01)
        return result


class NestedFuture(asyncio.Future):
    def __init__(self, is_async, func, args, kwargs, loop, executor=None):
        super().__init__(loop=loop)
        self._task = None
        self._start_cancelled = False

        if is_async:
            self._cf = partial(func, *args, **kwargs)
        else:
            f = partial(func, *args, **kwargs)
            self._cf = partial(self._run_in_executor, loop, executor, f)

        self.func = func
        self.args = args
        self.kwargs = kwargs
        self.executor = executor

    @staticmethod
    async def _run_in_executor(loop, executor, func):
        future = loop.run_in_executor(executor, func)
        return await future

    def start(self):
        if self.cancelled():
            return

        if self._task:
            return

        task = self._loop.create_task(self._cf())
        self._task = task
        task.add_done_callback(self._complete)

    def _complete(self, task):
        try:
            result = task.result()
            self.set_result(result)
        except BaseException as e:
            self.set_exception(e)

    @property
    def is_running(self):
        if self._task is None:
            return False

        return not self._task.done()

    def cancel(self):
        self._start_cancelled = True

        if self._task:
            self._task.cancel()
        else:
            self._super_cancel()

    def _super_cancel(self):
        super().cancel()

    def __str__(self):
        info = {
            "func": self.func.__name__,
            "args": self.args,
            "kwargs": self.kwargs,
            "state": self._state,
        }
        return f"{info}"


class AsyncPoolExecutor(PExecutor, PExecutable):
    def __init__(self, max_workers: int):
        self._max_workers = max_workers
        self._tasks = set()
        self._pre_queue = deque()
        self.is_running = False
        self._is_closed = False

        self._loop: asyncio.BaseEventLoop = None  # type: ignore
        self._threadpool: _ThreadPoolExecutor = None  # type: ignore
        self._consumer: asyncio.Task = None  # type: ignore

    def __executor__(self) -> PExecutor:
        return self

    @property
    def is_closed(self):
        return self._is_closed

    def __enter__(self):
        raise NotImplementedError()

    def __exit__(self, *args, **kwargs):
        raise NotImplementedError()

    async def __aenter__(self) -> PExecutor:
        if self._loop:
            raise RuntimeError("already initialized.")

        with ExitStack() as stack:
            self._loop = asyncio.get_event_loop()  # type: ignore
            self._threadpool = stack.enter_context(
                _ThreadPoolExecutor(self.max_workers)
            )
            self._consumer = self._loop.create_task(self.consume())
            stack.pop_all()

        self.is_running = True
        return self

    async def __aexit__(self, *args, **kwargs):
        await self.aclose()

    async def aclose(self):
        self._is_closed = True
        tasks = self._tasks

        # キューイングされた全てのタスクがスケジュールされるまで待機する
        while len(self._pre_queue):
            await asyncio.sleep(0.05)

        self._consumer.cancel()

        # 実行中のタスクを完了まで待機する
        while len(tasks):
            await asyncio.sleep(0.05)

        self._threadpool.shutdown()
        self.is_running = False

    @property
    def is_full(self):
        exists = self.max_workers <= len(self._tasks)
        return exists

    def _add_future(self, future):
        self._tasks.add(future)

    def _remove_future(self, future):
        try:
            self._tasks.remove(future)
        except Exception:
            pass

    def append_to_queue(self, item):
        self._pre_queue.append(item)

    async def consume(self):
        try:
            queue = self._pre_queue

            while True:
                while self.is_full:
                    if not self.is_running:
                        break
                    await asyncio.sleep(0.05)

                if not self.is_running:
                    break

                while True:
                    try:
                        future: NestedFuture = queue.popleft()
                        break
                    except IndexError:
                        await asyncio.sleep(0.05)

                future.start()
                self._add_future(future)
                future.add_done_callback(self._remove_future)
        except asyncio.CancelledError:
            ...
        except BaseException as e:
            logging.exception(str(e))
            raise

    def _asubmit_inner(self, is_async, func, args=None, kwargs=None):
        args = args or tuple()
        kwargs = kwargs or {}
        if self.is_closed:
            raise RuntimeError("cannot schedule new futures after shutdown")

        future = NestedFuture(
            is_async, func, args, kwargs, self._loop, self._threadpool
        )
        self.append_to_queue(future)
        return future

    def submit(self, func, *args, **kwargs):
        raise NotImplementedError()

    def asubmit(self, func, *args, **kwargs):
        is_async = iscoroutinefunction(func)
        return self._asubmit_inner(is_async, func, args, kwargs)

    def amap(self, func, iterable):
        return self._amap(func, iterable)

    def _amap(self, func, iterable):
        if isinstance(func, partial):
            target = func.args[0]
        else:
            target = func

        if asyncio.iscoroutinefunction(target):

            async def main():
                return await asyncio.gather(*(func(x) for x in iterable))

            return asyncio.create_task(main())
        else:
            return self._asubmit_inner(False, map, (func, iterable))

    @property
    def running_task_count(self):
        return len(self._tasks)

    def map(self, iterable, func, unpack="", timeout=None):
        return tools.map(self, iterable, func, unpack, timeout=timeout)

    @property
    def is_async_only(self):
        return True

    @property
    def max_workers(self):
        return self._max_workers

    @property
    def is_cpubound(self):
        return False


async def exec_func(loop, func):
    return await loop.run_in_executor(None, func)


async def exec_func_async(sem, func):
    while sem._value == 0:
        await asyncio.sleep(0)

    with sem:
        return await func()


class DummyPoolExecutor(OverrideExecutor):
    def __init__(self, limit):
        import threading

        self.max_workers = limit
        self._shutdown = False
        self._sem = threading.Semaphore(limit)

    def __executor__(self) -> PExecutor:
        return self  # type: ignore

    def submit(self, func, *args, **kwargs):
        if self._shutdown:
            raise RuntimeError("cannot schedule new futures after shutdown")

        f = Future()

        try:
            if iscoroutinefunction(func):
                result = asyncio.run(func(*args, **kwargs))
            else:
                result = func(*args, **kwargs)

        except BaseException as e:
            f.set_exception(e)
        else:
            f.set_result(result)

        return f

    def asubmit(self, func, *args, **kwargs):
        if self._shutdown:
            raise RuntimeError("cannot schedule new futures after shutdown")

        if iscoroutinefunction(func):
            async_func = partial(func, *args, **kwargs)
        else:
            async_func = partial(
                exec_func, asyncio.get_running_loop(), partial(func, *args, **kwargs)
            )
        return asyncio.create_task(exec_func_async(self._sem, async_func))

    def shutdown(self, wait=True):
        self._shutdown = True

    def __enter__(self):
        return self

    def __exit__(self, *args, **kwargs):
        ...

    async def __aenter__(self):
        return self

    async def __aexit__(self, *args, **kwargs):
        ...

    @property
    def is_cpubound(self):
        return False

    @property
    def is_async_only(self):
        return False


defaultpool = None


def get_default_pool():
    global defaultpool
    if defaultpool is None:
        defaultpool = DummyPoolExecutor(10)

    return defaultpool



from typing import Any, Callable, Coroutine, TypeVar

from pnq.aio import run

from .core import Query, QueryAsync, QueryDict, QueryNormal, QuerySeq, QuerySet

R = TypeVar("R")


class Builder:
    QUERY_BOTH = Query
    QUERY_ASYNC = QueryAsync
    QUERY_NORMAL = QueryNormal
    QUERY_SEQ = QuerySeq
    QUERY_DICT = QueryDict
    QUERY_SET = QuerySet

    @classmethod
    def query(cls, source):
        if isinstance(source, list):
            return cls.QUERY_SEQ(source)
        elif isinstance(source, dict):
            return cls.QUERY_DICT(source)
        elif isinstance(source, tuple):
            return cls.QUERY_SEQ(source)
        elif isinstance(source, set):
            return cls.QUERY_SET(source)
        elif isinstance(source, frozenset):
            return cls.QUERY_SET(source)
        elif isinstance(source, Query):
            return source
        else:
            has_iter = hasattr(source, "__iter__")
            has_aiter = hasattr(source, "__aiter__")

            if has_iter and has_aiter:
                return cls.QUERY_BOTH(source)
            elif has_iter:
                return cls.QUERY_NORMAL(source)
            elif has_aiter:
                return cls.QUERY_ASYNC(source)
            else:
                raise TypeError()

    def infinite(func, *args, **kwargs):
        def infinite(*args, **kwargs):
            while True:
                yield func(*args, **kwargs)

        return query(LazyGenerator(infinite, *args, **kwargs))

    def count(start=0, step=1):
        from itertools import count

        return query(LazyGenerator(count, start, step))

    def cycle(iterable):
        from itertools import cycle

        return query(LazyGenerator(cycle, iterable))

    @classmethod
    def run(
        cls,
        func: Callable[..., Coroutine[Any, Any, R]],
        handle_signals={"SIGINT", "SIGTERM"},
    ) -> R:
        return run(func, handle_signals)



import asyncio
import random
from typing import Iterable, Tuple, TypeVar

from pnq import selectors
from pnq._itertools.common import Listable, name_as
from pnq.exceptions import DuplicateElementError, MustError, MustTypeError

from . import finalizers

T = TypeVar("T")
K = TypeVar("K")
V1 = TypeVar("V1")
V2 = TypeVar("V2")


@name_as("map")
def _map(source: Iterable[T], selector, unpack=""):
    if selector is None:
        return source.__iter__()
    else:
        if selector is str:
            selector = lambda x: "" if x is None else str(x)

        if unpack == "":
            return (selector(x) for x in source)
        elif unpack == "*":
            return (selector(*x) for x in source)
        elif unpack == "**":
            return (selector(**x) for x in source)
        elif unpack == "***":
            return (selector(*x.args, **x.kwargs) for x in source)  # type: ignore
        else:
            raise ValueError("unpack must be one of '', '*', '**', '***'")


def flat(source: Iterable[T], selector=None):
    if selector is None:
        return (_ for inner in source for _ in inner)
    else:
        return (_ for elm in source for _ in selector(elm))


def traverse(source: Iterable[T], selector):
    scanner = selectors.traverse(selector)
    for root in source:
        for x in scanner(root):
            yield x


def pivot_unstack(source: Iterable[T], default=None):
    dataframe = {}  # type: ignore
    data = []

    # 全てのカラムを取得
    for i, dic in _enumerate(source):
        data.append(dic)
        for k in dic.keys():
            dataframe[k] = None

    # カラム分の領域を初期化
    for k in dataframe:
        dataframe[k] = []

    # データをデータフレームに収める
    for dic in data:
        for k in dataframe.keys():
            v = dic.get(k, default)
            dataframe[k].append(v)

    for x in dataframe.items():
        yield x


def pivot_stack(source: Iterable[T], default=None):
    data = dict(Listable(source))
    columns = list(data.keys())

    for i in range(len(columns)):
        row = {}
        for c in columns:
            row[c] = data[c][i]

        yield row


@name_as("enumerate")
def _enumerate(source: Iterable[T], start: int = 0, step: int = 1):
    i = start
    for x in source:
        yield i, x
        i += step


def group_by(source: Iterable[T], selector):
    from collections import defaultdict

    results = defaultdict(list)
    for elm in source:
        k, v = selector(elm)
        results[k].append(v)

    for k, v in results.items():
        yield k, v


def chain(*iterables):

    prev = 3

    for iterable in iterables:
        type = 0
        if hasattr(iterable, "__iter__"):
            type += 1
        if hasattr(iterable, "__iter__"):
            type += 2

        if type == 1:
            prev = 1
            may_be_async = False
        elif type == 2:
            prev = 2
            may_be_async = True
        elif type == 3:
            prev = 3
            if prev == 1:
                may_be_async = False
            elif prev == 2:
                may_be_async = True
            elif prev == 3:
                may_be_async = True
            else:
                raise TypeError(iterable, "iterable")
        else:
            raise TypeError(iterable, "iterable")

        if may_be_async:
            for x in iterable:
                yield x
        else:
            for x in iterable:
                yield x


def chunk(source: Iterable[T], size: int):
    current = 0
    running = True

    it = source.__iter__()

    while running:
        current = 0
        queue = []

        while current < size:
            current += 1
            try:
                val = it.__next__()
                queue.append(val)
            except StopIteration:
                running = False
                break

        if queue:
            yield queue


def tee(source: Iterable[T], size: int):
    ...


def join(source: Iterable[T], size: int):
    [].join(
        [],
        lambda left, right: left.name == right.name,
        lambda left, right: (left.name, right.age),
    )

    table(User).join(Item, on=User.id == Item.id).select(User.id, Item.id)

    pass


def inner_join(
    source: Iterable[Tuple[K, V1]],
    other: Iterable[Tuple[K, V2]],
):
    """
    Implementation for part of join_impl
    :param other: other sequence to join with
    :param sequence: first sequence to join with
    :return: joined sequence
    """
    seq_kv = finalizers.to_dict(Listable(source))
    other_kv = finalizers.to_dict(Listable(other))

    keys = seq_kv.keys() if len(seq_kv) < len(other_kv) else other_kv.keys()
    result = {}
    for k in keys:
        if k in seq_kv and k in other_kv:
            result[k] = (seq_kv[k], other_kv[k])
    for k, v in result.items():
        yield k, v


def group_join(source: Iterable[T], size: int):
    ...


def debug(source: Iterable[T], breakpoint=lambda x: x, printer=print):
    for v in source:
        printer(v)
        breakpoint(v)
        yield v


def union_all(source: Iterable[T]):
    ...


extend = union_all


def union(source: Iterable[T]):
    ...


def union_intersect(source: Iterable[T]):
    ...


def union_minus(source: Iterable[T]):
    ...


# difference
# symmetric_difference

# < <= > >= inclusion


@name_as("zip")
def _zip(source: Iterable[T]):
    ...


def compress(source: Iterable[T]):
    import itertools

    return itertools.compress(self, *iterables)


def cartesian(*iterables: Iterable[T]):
    # FIXME: no implemented
    import itertools

    for x in itertools.product(*iterables):  # type: ignore
        yield x


@name_as("filter")
def _filter(source: Iterable[T], predicate=None, unpack=""):
    if predicate is None:
        return source.__iter__()
    else:
        if unpack == "":
            return (x for x in source if predicate(x))
        elif unpack == "*":
            return (x for x in source if predicate(*x))
        elif unpack == "**":
            return (x for x in source if predicate(**x))
        elif unpack == "***":
            return (x for x in source if predicate(*x.args, **x.kwargs))  # type: ignore
        else:
            raise ValueError("unpack must be one of '', '*', '**', '***'")


def must(source: Iterable[T], predicate, msg=""):
    for elm in source:
        if not predicate(elm):
            raise MustError(f"{msg} {elm}")
        yield elm


def filter_type(source: Iterable[T], *types):
    for elm in source:
        if isinstance(elm, *types):
            yield elm


def must_type(source: Iterable[T], *types):
    for elm in source:
        if not isinstance(elm, types):
            raise MustTypeError(f"{elm} is not {types}")
        yield elm


def filter_keys(source: Iterable[T], *keys):
    pass


def must_keys(source: Iterable[T], *keys):
    pass


def filter_unique(source: Iterable[T], selector=None):
    duplidate = set()
    for value in _map(source, selector):
        if value in duplidate:
            pass
        else:
            duplidate.add(value)
            yield value


def must_unique(source: Iterable[T], selector=None):
    duplidate = set()
    for value in _map(source, selector):
        if value in duplidate:
            raise DuplicateElementError(value)
        else:
            duplidate.add(value)
            yield value


def take(source: Iterable[T], range: range):
    start = range.start
    stop = range.stop
    step = range.step

    current = 0

    it = source.__iter__()

    try:
        while current < start:
            it.__next__()
            current += step
    except StopIteration:
        return

    try:
        while current < stop:
            yield it.__next__()
            current += step
    except StopIteration:
        return


def take_while(source: Iterable[T], predicate):
    for v in source:
        if predicate(v):
            yield v
        else:
            break


def skip(source: Iterable[T], range: range):
    start = range.start
    stop = range.stop

    current = 0

    it = source.__iter__()

    try:
        while current < start:
            yield it.__next__()
            current += 1
    except StopIteration:
        return

    try:
        while current < stop:
            it.__next__()
            current += 1
    except StopIteration:
        return

    for x in it:
        yield x


def skip_while(source: Iterable[T], predicate):
    for v in source:
        if not predicate(v):
            break

    for v in source:
        yield v


def take_page(source: Iterable[T], page: int, size: int):
    r = _take_page_calc(page, size)
    return take(source, r)


def _take_page_calc(page: int, size: int):
    if page < 1:
        raise ValueError("page must be greater than 0")
    if size < 0:
        raise ValueError("size must be >= 0")
    start = (page - 1) * size
    stop = start + size
    return range(start, stop)


def order_by(source: Iterable[T], key_selector=None, desc: bool = False):
    for x in sorted(Listable(source), key=key_selector, reverse=desc):
        yield x


def min_by(source: Iterable[T], key_selector=None):
    for x in sorted(Listable(source), key=key_selector, reverse=False):
        yield x


def max_by(source: Iterable[T], key_selector=None):
    for x in sorted(Listable(source), key=key_selector, reverse=True):
        yield x


def order_by_reverse(source: Iterable[T]):
    return reversed(Listable(source))


def order_by_shuffle(source: Iterable[T], seed=None):
    if seed is None:
        rand = random.random
        func = lambda x: rand()  # noqa

    else:
        rand = random.Random(seed).random
        func = lambda x: rand()  # noqa

    for x in sorted(Listable(source), key=func):
        yield x


def sleep(source: Iterable[T], seconds: float):
    sleep = asyncio.sleep
    for v in source:
        yield v
        sleep(seconds)


def as_aiter(source):
    type = 0
    if hasattr(source, "__iter__"):
        type += 1
    if hasattr(source, "__iter__"):
        type += 2

    if type == 3:
        for x in source:
            yield x
    elif type == 2:
        for x in source:
            yield x
    elif type == 1:
        for x in source:
            yield x
    else:
        raise TypeError(f"{source} must be iterable")



class PnqException(Exception):
    """
    Pnqに関連する全ての例外の基底クラス
    """

    def __init__(self, msg: str = ""):
        super().__init__(msg)


class MustError(PnqException):
    """
    クエリがシーケンス全体あるいはシーケンスの要素に対して任意の特性を持っていることを要求したが失敗した

    関連: `must`
    """


class MustTypeError(PnqException, TypeError):
    """
    クエリがシーケンスの要素に対して任意の型であることを要求したが失敗した

    関連: `must_type`
    """


# KeyError IndexErrorを継承しても try: ... except KeyError としても補足できないっぽい
class NotFoundError(MustError, KeyError, IndexError):
    """
    クエリがキーに対応する要素を要求したが存在しない。
    IndexErrorとKeyErrorはKeyNotFoundErrorに置き換わります

    関連: `get` `must_get_many`
    """

    def __init__(self, key):
        super().__init__(f"{key}")
        self.key = key


class NoElementError(MustError):
    """
    クエリが何らかの要素を要求したが要素が存在しない

    関連: `one` `first` `last`
    """

    pass


class NotOneElementError(MustError):
    """
    クエリが要素がひとつであることを要求したが複数の要素が存在した

    関連: `one`
    """

    pass


class DuplicateElementError(MustError):
    """
    クエリが要素が重複していないことを要求したが重複を検知した

    関連: `must_unique`
    """
